import { defer } from 'rxjs/observable/defer';
import { merge } from 'rxjs/observable/merge';
import { filter, startWith, switchMap, take, takeUntil } from 'rxjs/operators';
import { Subject } from 'rxjs/Subject';
import { transition, trigger } from '@angular/animations';
import { ActiveDescendantKeyManager } from '@angular/cdk/a11y';
import { coerceBooleanProperty } from '@angular/cdk/coercion';
import { SelectionModel } from '@angular/cdk/collections';
import { DOWN_ARROW, END, ENTER, HOME, LEFT_ARROW, RIGHT_ARROW, SPACE, UP_ARROW } from '@angular/cdk/keycodes';
import { CdkOverlayOrigin } from '@angular/cdk/overlay';
import { ChangeDetectorRef, Component, ContentChildren, ElementRef, EventEmitter, Input, isDevMode, NgZone, Optional, Output, Self, ViewChild, ViewEncapsulation, NgModule } from '@angular/core';
import { NgControl } from '@angular/forms';
import { fadeIn, fadeOut } from '@ng-tangram/animate/fading';
import { NT_OPTION_PARENT_COMPONENT, NtOptionComponent, NtOverlayComponent, NtOptionModule, NtOverlayModule } from '@ng-tangram/components/core';
import { NtFormFieldControl } from '@ng-tangram/components/forms';
import { CommonModule } from '@angular/common';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
/**
 * @return {?}
 */
function getNtSelectDynamicMultipleError() {
    return Error('Cannot change `multiple` mode of select after initialization.');
}
/**
 * @return {?}
 */
function getNtSelectNonArrayValueError() {
    return Error('Value must be an array in multiple-selection mode.');
}
/**
 * @return {?}
 */
function getNtSelectNonFunctionValueError() {
    return Error('`compareWith` must be a function.');
}
class NtSelectChange {
    /**
     * @param {?} source
     * @param {?} value
     */
    constructor(source, value) {
        this.source = source;
        this.value = value;
    }
}
class NtSelectComponent extends NtFormFieldControl {
    /**
     * @param {?} _ngZone
     * @param {?} _elementRef
     * @param {?} _changeDetectorRef
     * @param {?} ngControl
     */
    constructor(_ngZone, _elementRef, _changeDetectorRef, ngControl) {
        super();
        this._ngZone = _ngZone;
        this._elementRef = _elementRef;
        this._changeDetectorRef = _changeDetectorRef;
        this.ngControl = ngControl;
        this._destroy = new Subject();
        this._disabled = false;
        this._focused = false;
        this._state = '';
        this._placeholder = '';
        this._multiple = false;
        this._required = false;
        this._compareWith = (o1, o2) => o1 === o2;
        this._onChange = () => { };
        this._onTouched = () => { };
        this.filterNotFound = 'Not Found';
        this.afterOpen = new EventEmitter();
        this.afterClosed = new EventEmitter();
        this.beforeOpen = new EventEmitter();
        this.beforeClosed = new EventEmitter();
        this.positionChange = new EventEmitter();
        this.selectionChange = new EventEmitter();
        this.valueChange = new EventEmitter();
        this.optionSelectionChanges = defer(() => {
            if (this.options) {
                return merge(...this.options.map(option => option.selectionChange));
            }
            return this._ngZone.onStable
                .asObservable()
                .pipe(take(1), switchMap(() => this.optionSelectionChanges));
        });
        this.origin = new CdkOverlayOrigin(this._elementRef);
        if (this.ngControl) {
            this.ngControl.valueAccessor = this;
        }
    }
    /**
     * @return {?}
     */
    get value() { return this._value; }
    /**
     * @return {?}
     */
    get triggerValue() {
        if (this.empty) {
            return '';
        }
        if (this._multiple) {
            const /** @type {?} */ selectedOptions = this._selectionModel.selected.map(option => option.label);
            return selectedOptions.join(', ');
        }
        return this._selectionModel.selected[0].label;
    }
    /**
     * @return {?}
     */
    get empty() { return !this._selectionModel || this._selectionModel.isEmpty(); }
    /**
     * @return {?}
     */
    get optionEmpty() {
        return this.options ? this.options.filter(option => !option.hidden).length === 0 : false;
    }
    /**
     * @return {?}
     */
    get focused() { return this._focused; }
    /**
     * @param {?} value
     * @return {?}
     */
    set disabled(value) { this._disabled = coerceBooleanProperty(value); }
    /**
     * @return {?}
     */
    get disabled() { return this._disabled; }
    /**
     * @return {?}
     */
    get required() { return this._required; }
    /**
     * @param {?} value
     * @return {?}
     */
    set required(value) { this._required = coerceBooleanProperty(value); }
    /**
     * @return {?}
     */
    get state() { return this._state; }
    /**
     * @return {?}
     */
    get width() { return this._width; }
    /**
     * @param {?} value
     * @return {?}
     */
    set multiple(value) { this._multiple = coerceBooleanProperty(value); }
    /**
     * @return {?}
     */
    get multiple() { return this._multiple; }
    /**
     * @param {?} value
     * @return {?}
     */
    set filter(value) {
        if (typeof value === 'function') {
            this._filter = value;
        }
        else if (coerceBooleanProperty(value)) {
            this._filter = (keyword, option) => {
                const /** @type {?} */ regex = new RegExp(keyword.toLowerCase(), 'g');
                return regex.test(option.label.toLowerCase());
            };
        }
    }
    /**
     * @return {?}
     */
    get filter() { return this._filter; }
    /**
     * @return {?}
     */
    get selected() {
        return this.multiple ? this._selectionModel.selected : this._selectionModel.selected[0];
    }
    /**
     * @param {?} value
     * @return {?}
     */
    set placeholder(value) { this._placeholder = value; }
    /**
     * @return {?}
     */
    get placeholder() {
        if (this.empty) {
            return this._placeholder;
        }
        return this.filter ? this.triggerValue : this._placeholder;
    }
    /**
     * @param {?} fn
     * @return {?}
     */
    set compareWith(fn) {
        if (typeof fn !== 'function') {
            throw getNtSelectNonFunctionValueError();
        }
        else {
            this._compareWith = fn;
            if (this._selectionModel) {
                this._initializeSelection();
            }
        }
    }
    /**
     * @return {?}
     */
    ngOnInit() {
        this._selectionModel = new SelectionModel(this.multiple, undefined, false);
    }
    /**
     * @return {?}
     */
    ngAfterContentInit() {
        this._initKeyManager();
        this.options.changes.pipe(startWith(null), takeUntil(this._destroy)).subscribe(() => {
            this._resetOptions();
            this._initializeSelection();
        });
    }
    /**
     * @return {?}
     */
    onResize() {
        this._width = this.inputElement.nativeElement.clientWidth;
    }
    /**
     * @param {?} event
     * @return {?}
     */
    onSearch(event) {
        if (this.focused && this.filter && !this.disabled) {
            const /** @type {?} */ target = event.target;
            this.options.forEach(option => option.hidden = !this.filter(target.value, option));
        }
    }
    /**
     * @param {?} event
     * @return {?}
     */
    _beforeOpen(event) {
        this._state = 'folded';
        this.onResize();
        this._scrollActiveOptionIntoView();
        this._highlightCorrectOption();
        this.beforeOpen.next(event);
    }
    /**
     * @param {?} event
     * @return {?}
     */
    _beforeClosed(event) {
        this._state = 'closed';
        this._onTouched();
        this.beforeClosed.next(event);
    }
    /**
     * @param {?} event
     * @return {?}
     */
    _afterOpen(event) {
        this.afterOpen.next(event);
    }
    /**
     * @param {?} event
     * @return {?}
     */
    _afterClosed(event) {
        this._resetFilterResult();
        this.inputElement.nativeElement.blur();
        this.afterClosed.next(event);
    }
    /**
     * @param {?} change
     * @return {?}
     */
    _positionChange(change) {
        this.positionChange.next(change);
    }
    /**
     * @return {?}
     */
    onFocus() {
        if (!this.overlay.isOpen && !this._disabled) {
            this.overlay.show();
            this._focused = true;
        }
    }
    /**
     * @return {?}
     */
    onBlur() {
        this._focused = false;
    }
    /**
     * @return {?}
     */
    focus() {
        this.inputElement.nativeElement.focus();
    }
    /**
     * @param {?} value
     * @return {?}
     */
    writeValue(value) {
        if (this.options) {
            this._setSelectionByValue(value);
        }
    }
    /**
     * @param {?} fn
     * @return {?}
     */
    registerOnChange(fn) {
        this._onChange = fn;
    }
    /**
     * @param {?} fn
     * @return {?}
     */
    registerOnTouched(fn) {
        this._onTouched = fn;
    }
    /**
     * @return {?}
     */
    clear() {
        if (!this.disabled) {
            this._clearSelection();
            this._value = null;
            this._onChange(this.value);
            this.valueChange.next(this.value);
        }
    }
    /**
     * @param {?} isDisabled
     * @return {?}
     */
    setDisabledState(isDisabled) {
        this._disabled = isDisabled;
    }
    /**
     * Handles all keydown events on the select.
     * @param {?} event
     * @return {?}
     */
    _handleKeydown(event) {
        if (!this.disabled) {
            this.overlay.isOpen ? this._handleOpenKeydown(event) : this._handleClosedKeydown(event);
        }
    }
    /**
     * Handles keyboard events when the selected is open.
     * @param {?} event
     * @return {?}
     */
    _handleOpenKeydown(event) {
        const /** @type {?} */ keyCode = event.keyCode;
        const /** @type {?} */ isArrowKey = keyCode === DOWN_ARROW || keyCode === UP_ARROW;
        const /** @type {?} */ manager = this._keyManager;
        if (keyCode === HOME || keyCode === END) {
            event.preventDefault();
            keyCode === HOME ? manager.setFirstItemActive() : manager.setLastItemActive();
        }
        else if (isArrowKey && event.altKey) {
            event.preventDefault();
            this.overlay.hide();
        }
        else if ((keyCode === ENTER || keyCode === SPACE) && manager.activeItem) {
            event.preventDefault();
            manager.activeItem.selectViaInteraction();
            if (keyCode === ENTER && this.multiple) {
                this.overlay.hide();
            }
        }
        else {
            const /** @type {?} */ previouslyFocusedIndex = manager.activeItemIndex;
            manager.onKeydown(event);
            if (this._multiple && isArrowKey && event.shiftKey && manager.activeItem &&
                manager.activeItemIndex !== previouslyFocusedIndex) {
                manager.activeItem.selectViaInteraction();
            }
        }
    }
    /**
     * @param {?} event
     * @return {?}
     */
    _handleClosedKeydown(event) {
        const /** @type {?} */ keyCode = event.keyCode;
        const /** @type {?} */ isArrowKey = keyCode === DOWN_ARROW || keyCode === UP_ARROW ||
            keyCode === LEFT_ARROW || keyCode === RIGHT_ARROW;
        const /** @type {?} */ isOpenKey = keyCode === ENTER || keyCode === SPACE;
        // Open the select on ALT + arrow key to match the native <select>
        if (isOpenKey || ((this.multiple || event.altKey) && isArrowKey)) {
            event.preventDefault(); // prevents the page from scrolling down when pressing space
            this.overlay.show();
        }
        else if (!this.multiple) {
            this._keyManager.onKeydown(event);
        }
    }
    /**
     * @return {?}
     */
    _resetOptions() {
        const /** @type {?} */ changedOrDestroyed = merge(this.options.changes, this._destroy);
        this.optionSelectionChanges
            .pipe(takeUntil(changedOrDestroyed), filter(event => event.isUserInput))
            .subscribe(event => {
            this._onSelect(event.source);
            if (!this.multiple && this.overlay.isOpen) {
                this.overlay.hide();
            }
        });
        // Listen to changes in the internal state of the options and react accordingly.
        // Handles cases like the labels of the selected options changing.
        merge(...this.options.map(option => option.stateChanges))
            .pipe(takeUntil(changedOrDestroyed))
            .subscribe(() => {
            this._changeDetectorRef.markForCheck();
        });
        this._resetFilterResult();
    }
    /**
     * Sets up a key manager to listen to keyboard events on the overlay panel.
     * @return {?}
     */
    _initKeyManager() {
        this._keyManager = new ActiveDescendantKeyManager(this.options)
            .withTypeAhead()
            .withVerticalOrientation();
        this._keyManager.tabOut.pipe(takeUntil(this._destroy)).subscribe(() => {
            // Restore focus to the trigger before closing. Ensures that the focus
            // position won't be lost if the user got focus into the overlay.
            this.focus();
            this.overlay.hide();
            // this.inputElement.nativeElement.blur();
        });
        this._keyManager.change.pipe(takeUntil(this._destroy)).subscribe(() => {
            if (this.overlay.isOpen && this.paneElement) {
                this._scrollActiveOptionIntoView();
            }
            else if (!this.overlay.isOpen && !this.multiple && this._keyManager.activeItem) {
                this._keyManager.activeItem.selectViaInteraction();
            }
        });
    }
    /**
     * @return {?}
     */
    _resetFilterResult() {
        this.options.forEach(option => option.hidden = false);
    }
    /**
     * @return {?}
     */
    _initializeSelection() {
        Promise.resolve().then(() => {
            this._setSelectionByValue(this.ngControl ? this.ngControl.value : this.value);
        });
    }
    /**
     * @param {?} value
     * @param {?=} isUserInput
     * @return {?}
     */
    _setSelectionByValue(value, isUserInput = false) {
        if (this.multiple && value) {
            if (!Array.isArray(value)) {
                throw getNtSelectNonArrayValueError();
            }
            this._clearSelection();
            value.forEach((currentValue) => this._selectValue(currentValue, isUserInput));
            this._sortValues();
        }
        else {
            this._clearSelection();
            const /** @type {?} */ correspondingOption = this._selectValue(value, isUserInput);
            // Shift focus to the active item. Note that we shouldn't do this in multiple
            // mode, because we don't know what option the user interacted with last.
            if (correspondingOption) {
                this._keyManager.setActiveItem(this.options.toArray().indexOf(correspondingOption));
            }
        }
        this._changeDetectorRef.markForCheck();
    }
    /**
     * @param {?} value
     * @param {?=} isUserInput
     * @return {?}
     */
    _selectValue(value, isUserInput = false) {
        const /** @type {?} */ correspondingOption = this.options.find((option) => {
            try {
                return option.value != null && this._compareWith(option.value, value);
            }
            catch (/** @type {?} */ error) {
                if (isDevMode()) {
                    console.warn(error);
                }
                return false;
            }
        });
        if (correspondingOption) {
            isUserInput ? correspondingOption.selectViaInteraction() : correspondingOption.select();
            this._selectionModel.select(correspondingOption);
        }
        return correspondingOption;
    }
    /**
     * @param {?=} fallbackValue
     * @return {?}
     */
    _propagateChanges(fallbackValue) {
        let /** @type {?} */ valueToEmit = null;
        if (this.multiple) {
            valueToEmit = (/** @type {?} */ (this.selected)).map(option => option.value);
        }
        else {
            valueToEmit = this.selected ? (/** @type {?} */ (this.selected)).value : fallbackValue;
        }
        this._value = valueToEmit;
        this.valueChange.emit(valueToEmit);
        this._onChange(valueToEmit);
        this.selectionChange.emit(new NtSelectChange(this, valueToEmit));
        this._changeDetectorRef.markForCheck();
    }
    /**
     * @param {?=} skip
     * @return {?}
     */
    _clearSelection(skip) {
        this._selectionModel.clear();
        this.options.forEach(option => {
            if (option !== skip) {
                option.deselect();
            }
        });
    }
    /**
     * @return {?}
     */
    _sortValues() {
        if (this.multiple) {
            this._selectionModel.clear();
            this.options.forEach(option => {
                if (option.selected) {
                    this._selectionModel.select(option);
                }
            });
        }
    }
    /**
     * @param {?} option
     * @return {?}
     */
    _onSelect(option) {
        const /** @type {?} */ wasSelected = this._selectionModel.isSelected(option);
        if (this.multiple) {
            this._selectionModel.toggle(option);
            wasSelected ? option.deselect() : option.select();
            this._keyManager.setActiveItem(this.options.toArray().indexOf(option));
            this._sortValues();
        }
        else {
            this._clearSelection(option.value == null ? undefined : option);
            if (option.value == null) {
                this._propagateChanges(option.value);
            }
            else {
                this._selectionModel.select(option);
            }
        }
        if (wasSelected !== this._selectionModel.isSelected(option)) {
            this._propagateChanges();
        }
    }
    /**
     * Highlights the selected item. If no option is selected, it will highlight
     * the first item instead.
     * @return {?}
     */
    _highlightCorrectOption() {
        if (this._keyManager) {
            if (this.empty) {
                this._keyManager.setFirstItemActive();
            }
            else {
                const /** @type {?} */ selectIndex = this.options.toArray().indexOf(this._selectionModel.selected[0]);
                this._keyManager.setActiveItem(selectIndex);
            }
        }
    }
    /**
     * @return {?}
     */
    _scrollActiveOptionIntoView() {
        const /** @type {?} */ activeOption = this._keyManager.activeItem || this.options.first;
        const /** @type {?} */ panelScrollTop = this.paneElement.nativeElement.scrollTop;
        const /** @type {?} */ panelScrollBottom = panelScrollTop + this.paneElement.nativeElement.offsetHeight;
        if (panelScrollBottom === 0) {
            this.paneElement.nativeElement.scrollTop = activeOption.getOffsetTop();
        }
        else if (activeOption.getOffsetTop() < panelScrollTop) {
            // 当活动选项超出 panel 的顶部范围时，重新设置 panel 可视区域。
            this.paneElement.nativeElement.scrollTop = activeOption.getOffsetTop();
        }
        else if (activeOption.getOffsetTop() + activeOption.getOffsetHeight() >= panelScrollBottom) {
            // 当活动选项超出 panel 的底部范围时，重新设置 panel 可视区域。
            const /** @type {?} */ scrollOffset = this.paneElement.nativeElement.offsetHeight - activeOption.getOffsetHeight();
            this.paneElement.nativeElement.scrollTop = activeOption.getOffsetTop() - scrollOffset;
        }
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this._destroy.next();
        this._destroy.complete();
    }
}
NtSelectComponent.decorators = [
    { type: Component, args: [{
                selector: 'nt-select',
                template: "<input #inputElement type=\"text\" class=\"nt-has-symbol\" (focus)=\"onFocus()\" (blur)=\"onBlur()\" (window:resize)=\"onResize()\" (keyup)=\"onSearch($event)\" [readonly]=\"!filter\" [placeholder]=\"placeholder\" [disabled]=\"disabled\" [value]=\"filter && focused ? '' : triggerValue\"> <span class=\"nt-select-caret\" [@fade] *ngIf=\"!triggerValue || disabled\"></span> <span class=\"nt-select-clear\" (click)=\"clear()\" [@fade] *ngIf=\"triggerValue && !disabled\"></span> <nt-overlay [origin]=\"origin\" trigger=\"click\" position=\"bottomLeft\" fixed (afterOpen)=\"_afterOpen($event)\" (afterClosed)=\"_afterClosed($event)\" (beforeOpen)=\"_beforeOpen($event)\" (beforeClosed)=\"_beforeClosed($event)\" (positionChange)=\"_positionChange($event)\" (keydownEvents)=\"_handleKeydown($event)\"> <div #paneElement class=\"nt-select-pane\" [class.folded]=\"state === 'folded'\" [style.width.px]=\"width\"> <ng-content select=\"nt-option\"></ng-content> <div class=\"filter-empty\" *ngIf=\"optionEmpty\" (click)=\"overlay.hide()\"> {{ filterNotFound }} </div> </div> </nt-overlay> ",
                providers: [
                    { provide: NT_OPTION_PARENT_COMPONENT, useExisting: NtSelectComponent },
                    { provide: NtFormFieldControl, useExisting: NtSelectComponent }
                ],
                encapsulation: ViewEncapsulation.None,
                // changeDetection: ChangeDetectionStrategy.OnPush,
                animations: [
                    trigger('fade', [
                        transition('* => void', fadeOut(.15)),
                        transition('void => *', fadeIn(.15))
                    ])
                ],
                host: {
                    'class': 'nt-select nt-form-control nt-has-symbol',
                    '(resize)': 'onResize()',
                    '[class.focus]': 'overlay.isOpen'
                }
            },] },
];
/** @nocollapse */
NtSelectComponent.ctorParameters = () => [
    { type: NgZone, },
    { type: ElementRef, },
    { type: ChangeDetectorRef, },
    { type: NgControl, decorators: [{ type: Self }, { type: Optional },] },
];
NtSelectComponent.propDecorators = {
    "disabled": [{ type: Input },],
    "required": [{ type: Input },],
    "multiple": [{ type: Input },],
    "filter": [{ type: Input },],
    "filterNotFound": [{ type: Input },],
    "placeholder": [{ type: Input },],
    "inputElement": [{ type: ViewChild, args: ['inputElement',] },],
    "paneElement": [{ type: ViewChild, args: ['paneElement',] },],
    "overlay": [{ type: ViewChild, args: [NtOverlayComponent,] },],
    "options": [{ type: ContentChildren, args: [NtOptionComponent,] },],
    "afterOpen": [{ type: Output },],
    "afterClosed": [{ type: Output },],
    "beforeOpen": [{ type: Output },],
    "beforeClosed": [{ type: Output },],
    "positionChange": [{ type: Output },],
    "selectionChange": [{ type: Output },],
    "valueChange": [{ type: Output },],
    "compareWith": [{ type: Input },],
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */
class NtSelectModule {
}
NtSelectModule.decorators = [
    { type: NgModule, args: [{
                imports: [CommonModule, NtOverlayModule, NtOptionModule],
                exports: [NtSelectComponent, NtOptionModule],
                declarations: [NtSelectComponent]
            },] },
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes} checked by tsc
 */

export { NtSelectModule, getNtSelectDynamicMultipleError, getNtSelectNonArrayValueError, getNtSelectNonFunctionValueError, NtSelectChange, NtSelectComponent };
//# sourceMappingURL=select.js.map
